﻿using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Windows.Threading;

//====================================================
// Описание работы классов и методов исходника на:
// https://www.interestprograms.ru
// Исходные коды программ и игр
//====================================================

namespace WpfAppVectorsUnknownCoordinate
{
	/// <summary>
	/// Interaction logic for MainWindow.xaml
	/// </summary>
	public partial class MainWindow : Window
	{
		// Селектор движения по исходному или вычисленным векторам.
		// 0 - исходный вектор
		// 1 - уменьшение длины (откат) вектора
		// 2 - увеличение длины вектора
		private int _typeTrajectory = 0;
		private readonly Random _random = new();
		// Фигуры учавствовающие в анимации движения по векторам
		private readonly Shape[] shapes = new Shape[3];
		// Счетчик движения фигур для синхронизации анимаций 
		private int order = 0;


		public MainWindow()
		{
			InitializeComponent();

			// Массив фигур для движения по траекториям.
			shapes[0] = ellipse1;
			shapes[1] = ellipse2;
			shapes[2] = ellipse3;
		}

		private void Canvas1_SizeChanged(object sender, SizeChangedEventArgs e)
		{
			// Обозначение точки начала движения эллипсов.
			Canvas.SetLeft(ellipseCenter, canvas1.ActualWidth / 2);
			Canvas.SetTop(ellipseCenter, canvas1.ActualHeight / 2);

			// Корректировка размеров родителей фигур
			border1.Margin = new Thickness(0, ActualHeight / 6, 0, ActualHeight / 6);
			canvas1.Margin = new Thickness(border1.ActualWidth / 6, 0, border1.ActualWidth / 6, 0);

			// --- Инициализация координат эллипсов ---
			for (int i = 0; i < shapes.Length; i++)
			{
				Canvas.SetLeft(shapes[i], canvas1.ActualWidth / 2);
				Canvas.SetTop(shapes[i], canvas1.ActualHeight / 2);
			}
		}


		/// <summary>
		/// Вычисление неизвестной координаты
		/// </summary>
		private static (double x2, double y2) ComputeUnknowCoordinates(Shape shape, int trajectory, double startX, double startY, double toX, double toY)
		{
			// Данные для вычисления координаты X искомого вектора.
			double x = startX;
			double y = startY;
			double x1 = toX;
			double y1 = toY;
			// Неизвестная координата
			double x2;

			// Выбор направления масштабирования искомого вектора
			int k = 1;
			// 1 - уменьшение (откат) вектора
			// -1 - удлинение вектора
			if (trajectory == 2) k = -1;
			double y2 = toY - k * (shape.Width / 2) * Math.Sign(toY - startY);

			// Вычисление неизвестной координаты.
			x2 = (x1 - x) * (y2 - y) / (y1 - y) + x;

			return (x2, y2);
		}


		#region Управление анимациями движения по траекториям векторов

		/// <summary>
		/// Запуск анимаций
		/// </summary>
		private void Button_Click(object sender, RoutedEventArgs e)
		{
			if (sender is Button button)
			{
				switch (button.Content.ToString())
				{
					case "Запуск":
						StartAnimations();
						button.IsEnabled = false;
						break;

					case "Изменение конечной точки":
						_ = _typeTrajectory == 2 ? _typeTrajectory = 0 : _typeTrajectory++;

						if (_typeTrajectory == 0) Title = "Формальная траектория - исходный вектор направления";
						if (_typeTrajectory == 1) Title = "Движение до препятствия - уменьшение вектора (откат)";
						if (_typeTrajectory == 2) Title = "Скрытие элемента за горизонт - удлинение вектора";
						break;
				}
			}

		}


		/// <summary>
		/// Запуск анимации через определенное время.
		/// </summary>
		private void DelayAnimation()
		{
			DispatcherTimer timer = new() { Interval = TimeSpan.FromSeconds(1.5) };
			timer.Tick += (sender, args) =>
			{
				timer.Stop();

				StartAnimations();
			};

			timer.Start();
		}

		#endregion


		#region Методы анимаций

		/// <summary>
		///  Запуск одновременных анимаций
		/// </summary>
		private void StartAnimations()
		{
			for (int i = 0; i < shapes.Length; i++)
			{
				StartMotion(shapes[i]);
			}
		}

		/// <summary>
		/// Запуск движения отдельной фигуры
		/// </summary>
		/// <param name="shape">текущий эллипс</param>
		private void StartMotion(Shape shape)
		{
			// Начало движения от центра родителя.
			double startX = canvas1.ActualWidth / 2;
			double startY = canvas1.ActualHeight / 2;

			// Временные линия для обозначения вектора траектории.
			(Line line1, Line line2) = VectorDirectionLines(false, _typeTrajectory);

			// Задержка перед началом перемещения
			double beginTime = 0.5;

			// Случайный выбор конечных координат.
			double toX = _random.Next(0, (int)canvas1.ActualWidth);
			double toY = canvas1.ActualHeight;
			// Конечная точка может быть вверху.
			int r = _random.Next(0, 2);
			if (r % 2 == 0) toY = 0;

			// Чертим линию
			line1.X1 = startX;
			line1.Y1 = startY;
			line1.X2 = toX;
			line1.Y2 = toY;

			// Линия завершения траектории движения.
			line2.X1 = toX;
			line2.Y1 = toY;

			// Если включена коррекция завершения траектории.
			if (_typeTrajectory > 0)
			{
				(toX, toY) = ComputeUnknowCoordinates(shape, _typeTrajectory, startX, startY, toX, toY);
			}
			// Конечные точки линии коррекции траектории
			line2.X2 = toX;
			line2.Y2 = toY;

			// Продолжительность анимации движения
			double duration = 0.4;
			//
			DoubleAnimation xMotion = new()
			{
				To = toX,
				FillBehavior = FillBehavior.Stop,
				Duration = new Duration(TimeSpan.FromSeconds(duration)),
				BeginTime = TimeSpan.FromSeconds(beginTime)
			};
			DoubleAnimation yMotion = new()
			{
				To = toY,
				FillBehavior = xMotion.FillBehavior,
				Duration = xMotion.Duration,
				BeginTime = xMotion.BeginTime,
			};

			// Задержка в конечной точки для удобства визуализации.
			duration = 0.8;
			//
			DoubleAnimation xStop = new()
			{
				From = toX,
				To = toX,
				FillBehavior = FillBehavior.Stop,
				Duration = new Duration(TimeSpan.FromSeconds(duration))
			};
			DoubleAnimation yStop = new()
			{
				From = toY,
				To = toY,
				FillBehavior = xStop.FillBehavior,
				Duration = xStop.Duration,
			};

			xMotion.Completed += (sender, eArgs) =>
			{
				shape.BeginAnimation(Canvas.LeftProperty, xStop);
				shape.BeginAnimation(Canvas.TopProperty, yStop);
			};

			xStop.Completed += (sender, eArgs) =>
			{
				// Удаление временных линий
				VectorDirectionLines(true);

				// Случайное изменение размера текущей фигуры перед анимацией.
				shape.Width = shape.Height = _random.Next(8, (int)border1.Margin.Top * 2);

				// Позиционирование элемента будет происходить относительно его центра.
				shape.Margin = new Thickness(shape.Width / 2 * -1, shape.Width / 2 * -1, 0, 0);

				// Возврат фигуры на исходную позицию - в центр родителя.
				Canvas.SetLeft(shape, canvas1.ActualWidth / 2);
				Canvas.SetTop(shape, canvas1.ActualHeight / 2);

				// Синхронизация одновременного запуска нескольких элементов.
				order++;
				if (order == shapes.Length)
				{
					order = 0;

					// Запуск отложеной анимации.
					DelayAnimation();
				}
			};

			// Запуск анимации движения.
			shape.BeginAnimation(Canvas.LeftProperty, xMotion);
			shape.BeginAnimation(Canvas.TopProperty, yMotion);
		}


		/// <summary>
		/// Визуализация направления векторов при помощи линий.
		/// </summary>
		private (Line line1, Line line2) VectorDirectionLines(bool erase, int trajectory = 0)
		{
			// Временная линия для обозначения траектории
			Line line1 = new()
			{
				StrokeThickness = 1,
				Stroke = Brushes.DarkBlue,
				StrokeDashArray = new DoubleCollection(new double[] { 2, 2 }),
				Uid = "lineTemp",
			};
			Panel.SetZIndex(line1, -2);

			// Временная линия уменьшения или удлинения вектора
			Line line2 = new()
			{
				StrokeThickness = 3,
				Stroke = trajectory == 2 ? Brushes.DarkBlue : Brushes.Red,
				Uid = "lineTemp",
			};
			Panel.SetZIndex(line2, -1);

			if (erase == true)
			{
				// Удаляем помеченные элементы-линии
				foreach (UIElement e in canvas1.Children.OfType<UIElement>().ToList())
				{
					if (e.Uid.Equals("lineTemp") == true) canvas1.Children.Remove(e);
				}
			}
			else
			{
				// Добавляем линии в коллекцию родителя.
				canvas1.Children.Add(line1);
				canvas1.Children.Add(line2);
			}

			return (line1, line2);
		}


		#endregion

	}
}